/**
 * Copyright 2013-2015, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 *
 * @emails oncall+relay
 */

'use strict';

jest.dontMock('RelayTaskScheduler');

describe('RelayTaskScheduler', () => {
  var RelayTaskScheduler;
  var RelayTestUtils;
  var resolveImmediate;

  beforeEach(() => {
    jest.resetModuleRegistry();
    RelayTaskScheduler = require('RelayTaskScheduler');
    RelayTestUtils = require('RelayTestUtils');
    resolveImmediate = require('resolveImmediate');

    jest.addMatchers(RelayTestUtils.matchers);
  });

  describe('default scheduler', () => {
    it('resolves to undefined when no callbacks are supplied', () => {
      var mockFunction = jest.genMockFunction();
      RelayTaskScheduler.await().then(mockFunction);

      jest.runAllTimers();

      expect(mockFunction).toBeCalledWith(undefined);
    });

    it('immediately invokes tasks', () => {
      var mockFunction = jest.genMockFunction();
      RelayTaskScheduler.await(mockFunction);

      jest.runAllTimers();

      expect(mockFunction).toBeCalled();
    });

    it('invokes multiple awaited tasks in order', () => {
      var mockOrdering = [];

      RelayTaskScheduler.await(() => mockOrdering.push('foo'));
      RelayTaskScheduler.await(() => mockOrdering.push('bar'));
      RelayTaskScheduler.await(() => mockOrdering.push('baz'));

      jest.runAllTimers();

      expect(mockOrdering).toEqual(['foo', 'bar', 'baz']);
    });

    it('awaits tasks awaited by other tasks contiguously', () => {
      var mockOrdering = [];

      RelayTaskScheduler.await(() => {
        mockOrdering.push('foo');
        RelayTaskScheduler.await(() => mockOrdering.push('bar'));
      });
      // Although `baz` is enqueued before `bar`, `bar` should execute first.
      RelayTaskScheduler.await(() => mockOrdering.push('baz'));

      jest.runAllTimers();

      expect(mockOrdering).toEqual(['foo', 'bar', 'baz']);
    });

    it('resolves to the task\'s return value', () => {
      var mockFunction = jest.genMockFunction();
      RelayTaskScheduler.await(() => 42).then(mockFunction);

      jest.runAllTimers();

      expect(mockFunction).toBeCalledWith(42);
    });

    it('forwards return values for multiple callbacks', () => {
      var mockOrdering = [];

      RelayTaskScheduler.await(
        () => {
          mockOrdering.push('foo');
          return 'bar';
        },
        prevValue => {
          mockOrdering.push(prevValue);
          return 'baz';
        }
      ).then(
        returnValue => {
          mockOrdering.push(returnValue);
        }
      );

      jest.runAllTimers();

      expect(mockOrdering).toEqual(['foo', 'bar', 'baz']);
    });

    it('aborts and rejects if a callback throws', () => {
      var mockError = new Error('Expected error.');
      var mockCallback = jest.genMockFunction();
      var mockFailureCallback = jest.genMockFunction();

      RelayTaskScheduler.await(
        () => 'foo',
        () => { throw mockError; },
        mockCallback,
      ).catch(mockFailureCallback);

      jest.runAllTimers();

      expect(mockCallback).not.toBeCalled();
      expect(mockFailureCallback).toBeCalledWith(mockError);
    });

    it('does not affect next chain of callbacks after rejection', () => {
      var mockError = new Error('Expected error.');
      var mockCallback = jest.genMockFunction();
      var mockFailureCallback = jest.genMockFunction();
      var mockSuccessCallback = jest.genMockFunction();

      RelayTaskScheduler.await(
        () => { throw mockError; },
      ).catch(mockFailureCallback);

      RelayTaskScheduler.await(
        mockCallback,
      ).then(mockSuccessCallback);

      jest.runAllTimers();

      expect(mockFailureCallback).toBeCalledWith(mockError);
      expect(mockCallback).toBeCalled();
      expect(mockSuccessCallback).toBeCalled();
    });
  });

  describe('injected scheduler', () => {
    var mockScheduler;
    var mockTasks;

    beforeEach(() => {
      mockTasks = [];
      mockScheduler = function(executeTask) {
        resolveImmediate(() => mockTasks.push(executeTask));
      };
    });

    it('allows injection of a scheduler to defer task execution', () => {
      RelayTaskScheduler.injectScheduler(mockScheduler);

      var mockFunction = jest.genMockFunction();
      RelayTaskScheduler.await(mockFunction);

      jest.runAllTimers();

      expect(mockFunction).not.toBeCalled();
      expect(mockTasks.length).toBe(1);

      // Execute the task, which should not return anything.
      expect(mockTasks[0]()).toBe(undefined);
      expect(mockFunction).toBeCalled();
    });

    it('allows an injected scheduler to defer multiple tasks', () => {
      RelayTaskScheduler.injectScheduler(mockScheduler);

      var mockOrdering = [];

      RelayTaskScheduler.await(() => {
        mockOrdering.push('foo');
        RelayTaskScheduler.await(() => mockOrdering.push('bar'));
      });
      RelayTaskScheduler.await(() => mockOrdering.push('baz'));

      jest.runAllTimers();

      // Scheduler only sees one task at a time.
      expect(mockTasks.length).toBe(1);
      mockTasks[0]();
      expect(mockOrdering).toEqual(['foo']);

      // Scheduler only sees the next task after `resolveImmediate`.
      expect(mockTasks.length).toBe(1);

      jest.runAllTimers();

      expect(mockTasks.length).toBe(2);
      mockTasks[1]();
      expect(mockOrdering).toEqual(['foo', 'bar']);

      jest.runAllTimers();

      expect(mockTasks.length).toBe(3);
      mockTasks[2]();
      expect(mockOrdering).toEqual(['foo', 'bar', 'baz']);
    });

    it('throws if the same task is executed more than once', () => {
      RelayTaskScheduler.injectScheduler(mockScheduler);

      var mockFunction = jest.genMockFunction();
      RelayTaskScheduler.await(mockFunction);

      jest.runAllTimers();

      mockTasks[0]();

      expect(() => {
        mockTasks[0]();
      }).toFailInvariant(
        'RelayTaskScheduler: Tasks can only be executed once.'
      );
    });
  });
});
